use crate::model::SectionRef;
use crate::option_future::OptionFuture;
use crate::section_interface::SectionInterface;
use mpsc::error::SendError;
use std::{
    collections::VecDeque,
    mem::swap,
    sync::{
        atomic::{AtomicI32, Ordering},
        Arc,
    },
    time::Duration,
};
use thiserror::Error;
use tokio::{
    spawn,
    sync::mpsc,
    time::{delay_for, Instant},
};
use tracing::{debug, trace, trace_span, warn};

#[derive(Debug, Clone, PartialEq, Eq)]
pub struct RunHandle(i32);

#[derive(Debug)]
struct SectionRunnerInner {
    next_run_id: AtomicI32,
}

impl SectionRunnerInner {
    fn new() -> Self {
        Self {
            next_run_id: AtomicI32::new(1),
        }
    }
}

#[derive(Clone, Debug)]
enum RunnerMsg {
    Quit,
    QueueRun(RunHandle, SectionRef, Duration),
    CancelRun(RunHandle),
    CancelAll,
    Pause,
    Unpause,
}

#[derive(Clone, Debug, PartialEq)]
enum RunState {
    Waiting,
    Running {
        start_time: Instant,
    },
    Finished,
    Cancelled,
    Paused {
        start_time: Instant,
        pause_time: Instant,
    },
}

#[derive(Debug)]
struct SecRun {
    handle: RunHandle,
    section: SectionRef,
    duration: Duration,
    total_duration: Duration,
    state: RunState,
}

impl SecRun {
    fn new(handle: RunHandle, section: SectionRef, duration: Duration) -> Self {
        Self {
            handle,
            section,
            duration,
            total_duration: duration,
            state: RunState::Waiting,
        }
    }

    fn is_running(&self) -> bool {
        matches!(self.state, RunState::Running{..})
    }

    #[allow(dead_code)]
    fn is_paused(&self) -> bool {
        matches!(self.state, RunState::Paused{..})
    }
}

struct RunnerTask {
    interface: Arc<dyn SectionInterface + Sync>,
    msg_recv: mpsc::Receiver<RunnerMsg>,
    running: bool,
    delay_future: OptionFuture<tokio::time::Delay>,
    paused: bool,
}

impl RunnerTask {
    fn new(
        interface: Arc<dyn SectionInterface + Sync>,
        msg_recv: mpsc::Receiver<RunnerMsg>,
    ) -> Self {
        Self {
            interface,
            msg_recv,
            running: true,
            delay_future: None.into(),
            paused: false,
        }
    }

    fn start_run(&mut self, run: &mut SecRun) {
        use RunState::*;
        debug!(section_id = run.section.id, "starting running section");
        self.interface
            .set_section_state(run.section.interface_id, true);
        run.state = Running {
            start_time: Instant::now(),
        };
    }

    fn finish_run(&mut self, run: &mut SecRun) {
        if run.is_running() {
            debug!(section_id = run.section.id, "finished running section");
            self.interface
                .set_section_state(run.section.interface_id, false);
            run.state = RunState::Finished;
        } else {
            warn!(
                section_id = run.section.id,
                state = debug(&run.state),
                "cannot finish run which is not running"
            );
        }
    }

    fn cancel_run(&mut self, run: &mut SecRun) {
        if run.is_running() {
            debug!(section_id = run.section.id, "cancelling running section");
            self.interface
                .set_section_state(run.section.interface_id, false);
        }
        run.state = RunState::Cancelled;
    }

    fn pause_run(&mut self, run: &mut SecRun) {
        use RunState::*;
        match run.state {
            Running { start_time } => {
                debug!(section_id = run.section.id, "pausing running section");
                self.interface
                    .set_section_state(run.section.interface_id, false);
                run.state = Paused {
                    start_time,
                    pause_time: Instant::now(),
                };
            }
            Waiting => {
                debug!(section_id = run.section.id, "pausing waiting section");
                run.state = Paused {
                    start_time: Instant::now(),
                    pause_time: Instant::now(),
                };
            }
            Finished | Cancelled | Paused { .. } => {}
        }
    }

    fn unpause_run(&mut self, run: &mut SecRun) {
        use RunState::*;
        match run.state {
            Paused {
                start_time,
                pause_time,
            } => {
                debug!(section_id = run.section.id, "unpausing section");
                self.interface
                    .set_section_state(run.section.interface_id, true);
                run.state = Running {
                    start_time: Instant::now(),
                };
                let ran_for = pause_time - start_time;
                run.duration -= ran_for;
            }
            Waiting | Finished | Cancelled | Running { .. } => {
                warn!(
                    section_id = run.section.id,
                    state = debug(&run.state),
                    "can only unpause paused section"
                );
            }
        }
    }

    fn process_queue(&mut self, run_queue: &mut VecDeque<SecRun>) {
        use RunState::*;
        loop {
            let current_run = match run_queue.front_mut() {
                Some(current_run) => current_run,
                None => break,
            };
            let run_finished = match (&current_run.state, self.paused) {
                (Waiting, false) => {
                    self.start_run(current_run);
                    self.delay_future = Some(delay_for(current_run.duration)).into();
                    false
                }
                (Running { start_time }, false) => {
                    let time_to_finish = start_time.elapsed() >= current_run.duration;
                    if time_to_finish {
                        self.finish_run(current_run);
                        self.delay_future = None.into();
                    }
                    time_to_finish
                }
                (Paused { .. }, false) => {
                    self.unpause_run(current_run);
                    self.delay_future = Some(delay_for(current_run.duration)).into();
                    false
                }
                (Waiting, true) => {
                    self.pause_run(current_run);
                    false
                }
                (Running { .. }, true) => {
                    self.pause_run(current_run);
                    self.delay_future = None.into();
                    false
                }
                (Paused { .. }, true) => false,
                (Cancelled, _) | (Finished, _) => true,
            };

            if run_finished {
                run_queue.pop_front();
            } else {
                break;
            }
        }
    }

    fn handle_msg(&mut self, msg: Option<RunnerMsg>, run_queue: &mut VecDeque<SecRun>) {
        let msg = msg.expect("SectionRunner channel closed");
        use RunnerMsg::*;
        trace!(msg = debug(&msg), "runner_task recv");
        match msg {
            Quit => self.running = false,
            QueueRun(handle, section, duration) => {
                run_queue.push_back(SecRun::new(handle, section, duration));
            }
            CancelRun(handle) => {
                for run in run_queue {
                    if run.handle != handle {
                        continue;
                    }
                    trace!(handle = handle.0, "cancelling run by handle");
                    self.cancel_run(run);
                }
            }
            CancelAll => {
                let mut old_runs = VecDeque::new();
                swap(&mut old_runs, run_queue);
                trace!(count = old_runs.len(), "cancelling all runs");
                for mut run in old_runs {
                    self.cancel_run(&mut run);
                }
            }
            Pause => {
                self.paused = true;
            }
            Unpause => {
                self.paused = false;
            }
        }
    }

    async fn start(mut self) {
        let span = trace_span!("runner_task");
        let _enter = span.enter();

        let mut run_queue: VecDeque<SecRun> = VecDeque::new();

        while self.running {
            self.process_queue(&mut run_queue);
            let delay_done = || {
                trace!("delay done");
            };
            tokio::select! {
                msg = self.msg_recv.recv() => self.handle_msg(msg, &mut run_queue),
                _ = &mut self.delay_future, if self.delay_future.is_some() => delay_done()
            };
        }
    }
}

#[derive(Debug, Clone, Error)]
#[error("the SectionRunner channel is closed")]
pub struct ChannelClosed;

pub type Result<T, E = ChannelClosed> = std::result::Result<T, E>;

impl<T> From<SendError<T>> for ChannelClosed {
    fn from(_: SendError<T>) -> Self {
        Self
    }
}

#[derive(Clone, Debug)]
pub struct SectionRunner {
    inner: Arc<SectionRunnerInner>,
    msg_send: mpsc::Sender<RunnerMsg>,
}

#[allow(dead_code)]
impl SectionRunner {
    pub fn new(interface: Arc<dyn SectionInterface + Sync>) -> Self {
        let (msg_send, msg_recv) = mpsc::channel(8);
        spawn(RunnerTask::new(interface, msg_recv).start());
        Self {
            inner: Arc::new(SectionRunnerInner::new()),
            msg_send,
        }
    }

    pub async fn quit(&mut self) -> Result<()> {
        self.msg_send.send(RunnerMsg::Quit).await?;
        Ok(())
    }

    pub async fn queue_run(
        &mut self,
        section: SectionRef,
        duration: Duration,
    ) -> Result<RunHandle> {
        let run_id = self.inner.next_run_id.fetch_add(1, Ordering::Relaxed);
        let handle = RunHandle(run_id);
        self.msg_send
            .send(RunnerMsg::QueueRun(handle.clone(), section, duration))
            .await?;
        Ok(handle)
    }

    pub async fn cancel_run(&mut self, handle: RunHandle) -> Result<()> {
        self.msg_send.send(RunnerMsg::CancelRun(handle)).await?;
        Ok(())
    }

    pub async fn cancel_all(&mut self) -> Result<()> {
        self.msg_send.send(RunnerMsg::CancelAll).await?;
        Ok(())
    }

    pub async fn pause(&mut self) -> Result<()> {
        self.msg_send.send(RunnerMsg::Pause).await?;
        Ok(())
    }

    pub async fn unpause(&mut self) -> Result<()> {
        self.msg_send.send(RunnerMsg::Unpause).await?;
        Ok(())
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::section_interface::MockSectionInterface;
    use crate::{
        model::Section,
        trace_listeners::{EventListener, Filters, SpanFilters, SpanListener},
    };
    use tracing_subscriber::prelude::*;

    #[tokio::test]
    async fn test_quit() {
        let quit_msg = EventListener::new(
            Filters::new()
                .target("sprinklers_rs::section_runner")
                .message("runner_task recv")
                .field_value("msg", "Quit"),
        );
        let task_span = SpanListener::new(
            SpanFilters::new()
                .target("sprinklers_rs::section_runner")
                .name("runner_task"),
        );
        let subscriber = tracing_subscriber::registry()
            .with(quit_msg.clone())
            .with(task_span.clone());
        let _sub = tracing::subscriber::set_default(subscriber);

        let interface = MockSectionInterface::new(6);
        let mut runner = SectionRunner::new(Arc::new(interface));
        tokio::task::yield_now().await;
        runner.quit().await.unwrap();
        tokio::task::yield_now().await;

        assert_eq!(quit_msg.get_count(), 1);
        assert_eq!(task_span.get_exit_count(), 1);
    }

    fn make_sections_and_interface() -> (Vec<SectionRef>, Arc<MockSectionInterface>) {
        let interface = Arc::new(MockSectionInterface::new(2));
        let sections: Vec<SectionRef> = vec![
            Arc::new(Section {
                id: 1,
                name: "Section 1".into(),
                interface_id: 0,
            }),
            Arc::new(Section {
                id: 2,
                name: "Section 2".into(),
                interface_id: 1,
            }),
        ];
        (sections, interface)
    }

    fn assert_section_states(interface: &MockSectionInterface, states: &[bool]) {
        for (id, state) in states.iter().enumerate() {
            assert_eq!(
                interface.get_section_state(id as u32),
                *state,
                "section interface id {} did not match",
                id
            );
        }
    }

    async fn advance(dur: Duration) {
        // HACK: advance should really be enough, but we need another yield_now
        tokio::time::pause();
        tokio::time::advance(dur).await;
        tokio::task::yield_now().await;
        tokio::time::resume();
    }

    #[tokio::test]
    async fn test_queue() {
        let (sections, interface) = make_sections_and_interface();
        let mut runner = SectionRunner::new(interface.clone());

        assert_section_states(&interface, &[false, false]);

        // Queue single section, make sure it runs
        runner
            .queue_run(sections[0].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        tokio::task::yield_now().await;

        assert_section_states(&interface, &[true, false]);

        advance(Duration::from_secs(11)).await;

        assert_section_states(&interface, &[false, false]);

        // Queue two sections, make sure they run one at a time
        runner
            .queue_run(sections[1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        runner
            .queue_run(sections[0].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        tokio::task::yield_now().await;

        assert_section_states(&interface, &[false, true]);

        advance(Duration::from_secs(11)).await;

        assert_section_states(&interface, &[true, false]);

        advance(Duration::from_secs(10)).await;

        assert_section_states(&interface, &[false, false]);

        runner.quit().await.unwrap();
        tokio::task::yield_now().await;
    }

    #[tokio::test]
    async fn test_cancel_run() {
        let (sections, interface) = make_sections_and_interface();
        let mut runner = SectionRunner::new(interface.clone());

        let run1 = runner
            .queue_run(sections[1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        let _run2 = runner
            .queue_run(sections[0].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        let run3 = runner
            .queue_run(sections[1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        tokio::task::yield_now().await;

        assert_section_states(&interface, &[false, true]);

        runner.cancel_run(run1).await.unwrap();
        tokio::task::yield_now().await;

        assert_section_states(&interface, &[true, false]);

        runner.cancel_run(run3).await.unwrap();
        advance(Duration::from_secs(11)).await;

        assert_section_states(&interface, &[false, false]);

        runner.quit().await.unwrap();
        tokio::task::yield_now().await;
    }

    #[tokio::test]
    async fn test_cancel_all() {
        let (sections, interface) = make_sections_and_interface();
        let mut runner = SectionRunner::new(interface.clone());

        runner
            .queue_run(sections[1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        runner
            .queue_run(sections[0].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        runner
            .queue_run(sections[1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, true]);

        runner.cancel_all().await.unwrap();
        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, false]);

        runner.cancel_all().await.unwrap();
        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, false]);

        runner.quit().await.unwrap();
        tokio::task::yield_now().await;
    }

    #[tokio::test]
    async fn test_pause() {
        let (sections, interface) = make_sections_and_interface();
        let mut runner = SectionRunner::new(interface.clone());

        let _run1 = runner
            .queue_run(sections[1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        let run2 = runner
            .queue_run(sections[0].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        let _run3 = runner
            .queue_run(sections[1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, true]);

        runner.pause().await.unwrap();
        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, false]);

        advance(Duration::from_secs(10)).await;
        assert_section_states(&interface, &[false, false]);

        runner.unpause().await.unwrap();
        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, true]);

        advance(Duration::from_secs(8)).await;
        assert_section_states(&interface, &[false, true]);

        advance(Duration::from_secs(2)).await;
        assert_section_states(&interface, &[true, false]);

        runner.pause().await.unwrap();
        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, false]);

        // cancel paused run
        runner.cancel_run(run2).await.unwrap();
        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, false]);

        runner.unpause().await.unwrap();
        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, true]);

        advance(Duration::from_secs(11)).await;
        assert_section_states(&interface, &[false, false]);

        runner.quit().await.unwrap();
        tokio::task::yield_now().await;
    }
}
